# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

PARANOID_GET_STRING_CONST_SIGNATURE = [
    4294967295,
    33,
    7109453100751455733,
    28,
    -3808689974395783757,
    32,
    65535,
    16,
    -65536,
    0,
]
PARANOID_GET_STRING_ARGUMENTS = ["J"]
PARANOID_GET_STRING_RETURN_TYPE = "Ljava/lang/String;"
